//----------------------------------------------------------------------
// File:			ALTS.h
// Programmer:		David Mount
// Description:		Basic ALTS definitions
// Last modified:	06/17/2015 (Version 1.0)
//----------------------------------------------------------------------
// Copyright (c) 2015 University of Maryland and David Mount.  All
// Rights Reserved.
// 
// This software and related documentation is part of the Approximate
// Least-Trimmed Squares (ALTS) software.  This software is provided
// under the provisions of the GNU Public License (GPL). See the file
// ReadMe.txt for further information.
// 
// The University of Maryland (U.M.) and the authors make no
// representations about the suitability or fitness of this software for
// any purpose.  It is provided "as is" without express or implied
// warranty.
//----------------------------------------------------------------------
// History:
//	Revision 1.0  06/17/2015
//		Initial release
//----------------------------------------------------------------------

#ifndef ALTS_H
#define ALTS_H

#include <cstdlib>						// standard libs
#include <cstddef>						// standard defines
#include <cmath>						// math includes
#include <cassert>						// C assertion checking
#include <iostream>						// I/O streams
#include <iomanip>						// I/O manipulators
#include <fstream>						// file I/O
#include <string>						// STL string
#include <vector>						// STL vector

//----------------------------------------------------------------------
//	version info
//----------------------------------------------------------------------
#define ALTS_VERSION "1.0"				// ALTS version and information
#define ALTS_VERSION_CMT "(Initial release)"
#define ALTS_COPYRIGHT "David M. Mount"
#define ALTS_RELEASE_DATE "06/17/2015"

//----------------------------------------------------------------------
//	Output precision
//	The defined constant ALTS_OUTPUT_PRECISION stores the precision with
//	which fixed-point real output will be printed. This may be
//	overridden by giving the quantity at compilation time. Points are
//	printed at higher precision for possible future storage.
//----------------------------------------------------------------------
#define ALTS_OUTPUT_PRECISION 4
#define ALTS_POINT_PRECISION 10

//----------------------------------------------------------------------
//	Use biased variance
//	Normally, the LTS cost for h inliers is defined as the sample
//	standard deviation of the h closest residuals, that is,
//
//		sqrt((1/(h-1)) sum_{i=1}^h r[i]^2)
//
//	where r[i] is the i-th smallest absolute residual. For compatibility
//	with earlier versions of the program, the multiplier 1/(h-1) can be
//	replaced with 1/h, thus producing a (slightly) biased estimate for
//	the standard deviation.
//
//	To use the biased variance, uncomment the following line.
//----------------------------------------------------------------------
// #define ALTS_USE_BIASED_VARIANCE

//----------------------------------------------------------------------
// Special numeric values
// System-dependent numeric values can usually be found in limits.h (as
// LONG_MAX, INT_MAX) or in float.h (as DBL_MAX, FLT_MAX).
//
// Not all systems have these files, however.  If you are using such a
// system, you should set the preprocessor symbol ALTS_NO_LIMITS_H when
// compiling, and modify the statements below to generate the
// appropriate value. For practical purposes, this does not need to be
// the maximum double value. It is sufficient that it be at least as
// large than the maximum squared distance between between any two
// points.
//----------------------------------------------------------------------
#ifdef ALTS_NO_LIMITS_H					// limits.h unavailable
  #include <cvalues>					// replacement for limits.h
  const double ALTS_DBL_MAX = MAXDOUBLE; // insert maximum double
#else
  #include <climits>
  #include <cfloat>
  const double ALTS_DBL_MAX = DBL_MAX;
#endif

//----------------------------------------------------------------------
//	Basic constants and enumerations
//----------------------------------------------------------------------
namespace alts {

	enum OutputLev {					// output levels
		SILENT,							// no output
		RESULTS,						// output results
		PERF_STATS,						// print performance stats
		LEVEL_STATS,					// print per-level performance stats
		PRINT_POINTS,					// print the points
		// OTHERS...
		N_OUTPUT_LEVS					// number of output levels
	};

	enum Algorithm {					// algorithm types
		BRANCH_AND_BOUND,				// branch and bound
		FAST_LTS,						// Fast-LTS
		FULL_SEARCH_LTS,				// exhaustive search
		// OTHERS...
		N_ALGS							// number of algorithms
	};

	enum SrchPriority {					// B&B search priorities
		MIN_UPPER_BOUND,				// cell with min upper bound
		MIN_LOWER_BOUND,				// cell with min lower bound
		MIN_GAP,						// cell with min upper-lower gap
		OLDEST_CELL,					// cell with lowest cell number
		MAX_SAMPLES,					// cell with most samples
		MIN_SAMP_VARIANCE,				// cell with min variance of samples
		// OTHERS...
		N_PRIORITIES					// number of priority classes
	};

	enum PQOrder {						// B&B search priority method
		LEX_ORDER,						// lexicographic order
		CYCLIC_ORDER,					// cyclic order
		WEIGHTED_ORDER					// weighed random order
	};

	enum SplitMethod {					// B&B cell splitting method
		MIDPT_SPLIT,					// standard midpoint
		SL_MIDPT,						// sliding midpoint
		MEDIAN_SPLIT,					// median split
		// OTHERS...
		N_SPLITS						// number of splitting methods
	};

}

//----------------------------------------------------------------------
//	Vector Types:
//
//	Vector objects are complicated a bit by the fact that we distinguish
//	between independent and dependent variables. There is only one
//	dependent, which means that the dimension of the point space is 1
//	more than the number of independent variables. All vector objects
//	are dynamically allocated as STL vectors. Constructors for sets of
//	vector objects (e.g., SlopeSet and PointSet) require default
//	constructor. Thus, before using the elements of such a set it is
//	important to reserve enough space in each of the vector objects
//	before attempting to access the individual coordinates.
//	
//	Each plane is represented in dual space as vector (b0, ..., b{d-1}),
//	where the plane of is defined:
//
//		y = b0 + b1 * x1 + b2 * x2 + ... + b{d-1} * x{d-1}.
//
//	We store each point as a pair (y,x), where x is represented by a
//	homogeneous d-vector [1, x1, x2, ..., x{d-1}]. This makes it
//	convenient to represent a hyperplane as [b0, b1, ..., b{d-1}].
//
//		DVector	A d-vector of independent variables. This is the basic
//				type used for Indep, Plane, and Slope.
//		Indep	A (d-1)-vector of independent variables. We represent
//				this as a (homogeneous) d-vector (1, x1, ..., x{d-1}).
//				Note that the actual coordinates are indexed from 1 to
//				d-1.
//		Point	A point in d-dimensional (primal) space, consisting of
//				both a dependent variable (y) and a vector of
//				independent variables.
//		Plane	A (d-1)-dimensional hyperplane. Such a hyperplane is
//				represented as a d-vector, (b0, b1, ..., b{d-1}).
//----------------------------------------------------------------------

//----------------------------------------------------------------------
//	Slope:
//	The geometric branch and bound algorithm is based on the recursive
//	subdivision of a solution space. The natural choice is that of a
//	Plane, namely a point in d-dimensional dual space.  Instead, we will
//	save one dimension by storing only the components (b1, ..., b{d-1}),
//	which intuitively.  Our internal representation will actually be the
//	same as that of a Plane, but we will simply assume that the b0
//	coordinate is 0.
//----------------------------------------------------------------------

//----------------------------------------------------------------------
//	Basic definitions
//	Before using any of these, the variable altsGlobals must be
//	initialized.
//----------------------------------------------------------------------

namespace alts {
	typedef double	Scalar;				// arithmetic results on Scalar
	const Scalar SCALAR_EPS = 1E-7;
										// a very large value
	const Scalar SCALAR_INFTY = ALTS_DBL_MAX;

	class DVector;						// basic vector types
	class Indep;
	class Plane;
	class Slope;
	class Point;

	class Params;						// parameter storage
	class SrchParams;					// search parameter storage
	class SrchPriorList;				// B&B cell priority
	class LTSResult;					// result of LTS search
	class MSubsetEnum;					// m-element subset enumerator
	class SubsetEnum;					// subset enumerator

	class PointSet;						// set of points
	class SlopeSet;						// set of slopes

											// range types
	template <class T> class Range;			// range of values
	typedef Range<int> IntRange;			// range of integers [lo..hi-1]
	typedef Range<Scalar> ScalarRange;		// range of scalars

	template <class T> class VectorRange;	// range of vector type
	typedef VectorRange<Indep> IndepRange;	// range of independent
	typedef VectorRange<Plane> PlaneRange;	// range of hyperplanes
	typedef VectorRange<Slope> SlopeRange;	// range of slopes
}

//----------------------------------------------------------------------
//	Range of values
//	A range is an interval, given by a low value and high value. This is
//	very basic for now. This should not be used for vector types (those
//	inherited from DVector).
//----------------------------------------------------------------------
template <class T>
class alts::Range {
public:
	T	lo;								// low endpoint
	T	hi;								// high endpoint
public:
	Range() : lo(), hi() { }			// default constructor
										// construct from endpoints
	Range(const T& ll, const T& hh) : lo(ll), hi(hh) { }

	bool contains(const T& x) const		// is lo <= x <= hi?
	{ return !((x < lo) || (hi < x)); }
};

//----------------------------------------------------------------------
//	Range of vector type
//	This is a range for vector types. The constructor is given the
//	dimension.
//
//	We assume that T implements a copy constructor and a constructor
//	from the dimension. We also assume it implements "<=", 
//----------------------------------------------------------------------
template <class T>
class alts::VectorRange {
public:
	T	lo;								// low endpoint
	T	hi;								// high endpoint
public:
										// construct from dimension
	VectorRange(int d = 1) : lo(d), hi(d) { }
										// construct from dim & endpoints
	VectorRange(const T& ll, const T& hh)
		: lo(ll), hi(hh) { }

	bool contains(const T& x) const		// is lo <= x <= hi?
	{ return ((lo <= x) && (x <= hi)); }

	void expand(Scalar factor)			// expand by factor
	{
		int d = lo.getDim();			// dimension
		for (int j = 1; j < d; j++) {
			Scalar delta = (factor-1)*(hi[j] - lo[j])/2;
			lo[j] -= delta;
			hi[j] += delta;
		}
	}

};

//----------------------------------------------------------------------
//	DVector - A d-dimensional vector
//	The basic element for the other vector types. We rely on the default
//	constructor. The 0th coordinate often has special meaning, and so we
//	distinguish its value in the constructor. The methods shiftCopy*()
//	are used for copying the vector to standard C-style arrays, subject
//	to a translation of coordinates.
//----------------------------------------------------------------------

class alts::DVector : public std::vector<Scalar> {
public:
										// constructor
	DVector(int d = 1, Scalar x0 = 0, Scalar xx = 0);

	int getNIndep() const				// number of indep vars
	{ return int(size()-1); }

	int getDim() const					// dimension
	{ return int(size()); }

										// shift & copy to v[s..s+dim-1]
	void shiftCopyTo(Scalar* v, int s = 0) const;

										// shift & copy from v[s..s+dim-1]
	void shiftCopyFrom(Scalar* v, int s = 0);
};

//----------------------------------------------------------------------
//	Indep - A vector of independent vectors
//----------------------------------------------------------------------

class alts::Indep : public DVector {
public:
	Indep(int d = 1, Scalar base = 0)	// constructor from base value
		: DVector(d, 1, base) { }
};

//----------------------------------------------------------------------
//	Plane - A (d-1)-dimensional hyperplane
//----------------------------------------------------------------------

class alts::Plane : public DVector {
public:
										// constructor from base values
	Plane(int d = 1, Scalar b0 = 0, Scalar bx = 0)
		: DVector(d, b0, bx) { }
										// construct from intercept/slope
	Plane(Scalar b0, const Slope& b);
										// residual with respect to p
	Scalar residual(const Point& p) const;

	Slope getSlope() const;				// extract slope part
};

//----------------------------------------------------------------------
//	Slope - A (d-1)-dimensional hyperplane, without intercept
//----------------------------------------------------------------------

class alts::Slope : public DVector {
public:
	Slope(int d = 1, Scalar base = 0)	// constructor from base value
		: DVector(d, 0, base) { }
};

//----------------------------------------------------------------------
//	Point - A point, consisting of independent variables and dependent
//	variable.
//----------------------------------------------------------------------

class alts::Point {
private:
	Indep		x;						// independent variables
	Scalar		y;						// dependent variable
public:
										// constructor from base value
	Point(int d = 1, Scalar xbase = 0, Scalar ybase = 0)
		: x(d, xbase), y(ybase) { }
										// construct from parts
	Point(const Indep& xx, const Scalar& yy) : x(xx), y(yy) { }

	int getDim() const					// get dimension
		{ return x.getDim(); }

	Scalar getDep() const				// getters
		{  return y; }

	const Indep& getIndep() const
		{  return x; }

	Scalar& getDep()
		{  return y; }

	Indep& getIndep()
		{  return x; }

	void setDep(Scalar yy)				// setters
		{  y = yy; }

	void setIndep(Indep xx)
		{  x = xx; }
};

//----------------------------------------------------------------------
//	PointSet - A set of points
//	This is basically a C++ STL vector of points, but with a couple of
//	utilities, which are used for the numerical routines. These routines
//	copy points (either the independent or dependent parts) into a
//	standard C-style array of Scalar, but in which the arrays are
//	subscripted starting with 1, not 0.
//
//	One issue is that we cannot specify the dimension of the point sets
//	at the time of construction (since the STL vector only invokes the
//	default constructor for each element of the set). Thus, the
//	constructor builds an empty set, and the insert() method should
//	be used for inserting elements into the set. It is assumed that all
//	the points in the set have the same dimension.
//
//	The shift...To and shift...From methods copy members to/from a
//	C-style array.
//----------------------------------------------------------------------

class alts::PointSet : public std::vector<Point> {
public:
										// default constructor
	PointSet() : std::vector<Point>() { }

	PointSet(const PointSet& pts)		// copy constructor (never use!)
	{ assert(false); }

	void insert(const Point& p)			// insert a point
	{ push_back(p); }

	void shiftIndepTo(					// shift-copy indeps to M[s..][t..]
		Scalar**	M,
		int			s,
		int			t) const;

	void shiftIndepFrom(				// shift-copy indeps from M[s..][t..]
		Scalar**	M,
		int			s,
		int			t);

	void shiftDepTo(					// shift-copy deps to v[s..]
		Scalar*		v,
		int			s) const;

	void shiftDepFrom(					// shift-copy deps from v[s..]
		Scalar*		v,
		int			s);
};

//----------------------------------------------------------------------
//	SlopeSet - A set of slopes
//	This is basically a C++ STL vector of slopes.
//----------------------------------------------------------------------

class alts::SlopeSet : public std::vector<Slope> {
public:
										// default constructor
	SlopeSet() : std::vector<Slope>() { }

	void generateSamples(				// generate slope samples
		const Params&	params);		// parameters

	void insert(const Slope& b)			// insert a slope
	{ push_back(b); }

	SlopeRange getBoundingBox() const;	// get bounding box

	Scalar getVariance() const;			// get spatial variance of slopes

	void partition(						// partition about cutting plane
		int			cutDim,				// cutting dimension
		Scalar		cutVal,				// cutting value
		SlopeSet&	lowSide,			// slopes on low side (ret'd)
		SlopeSet&	highSide) const;	// slopes on high side (ret'd)
};

//----------------------------------------------------------------------
//	Iterators of various types
//----------------------------------------------------------------------

namespace alts {
										// iterators
	typedef DVector::iterator DVectorIter;
	typedef Indep::iterator IndepIter;
	typedef Plane::iterator PlaneIter;
	typedef Slope::iterator SlopeIter;
										// const iterators
	typedef DVector::const_iterator DVectorConstIter;
	typedef Indep::const_iterator IndepConstIter;
	typedef Plane::const_iterator PlaneConstIter;
	typedef Slope::const_iterator SlopeConstIter;

										// set iterators
	typedef PointSet::iterator PointSetIter;
	typedef SlopeSet::iterator SlopeSetIter;
										// const iterators
	typedef PointSet::const_iterator PointSetConstIter;
	typedef SlopeSet::const_iterator SlopeSetConstIter;
}

//----------------------------------------------------------------------
//	alts::LTSResult - Result of ALTS
//	The LTS cost is defined to be the standard deviation of the inlier 
//	residuals.
//
//	improveFit() is defined in ALTSnumeric.cpp. It returns false if no
//	actual change takes place in the plane of fit.
//----------------------------------------------------------------------

class alts::LTSResult {
private:
	Scalar		variance;				// variance of squared residuals
	Plane		b;						// plane of fit
	int			timeStamp;				// when was best last set?
public:
	LTSResult(int d = 1)				// construct from dimension
		: variance(-1), b(d), timeStamp(-1) { }

										// construct from parts
	LTSResult(Scalar v, const Plane& bb, int t = -1)
		: variance(v), b(bb), timeStamp(t) { }

	Scalar getVariance() const			// getters
		{  return variance; }

	Scalar getCost() const
		{ return sqrt(variance); }

	const Plane& getPlane() const
		{  return b; }

	Plane& getPlane()
		{  return b; }

	int getTimeStamp() const
		{  return timeStamp; }

	void setVariance(Scalar v)			// setters
		{  variance = v; }

	int setTimeStamp(int t)
		{  return timeStamp = t; }


	bool improveFit(					// locally improve LTS fit
		const PointSet&		pts,		// the points
		int					h);			// trimming parameter

										// get relative error wrt r
	Scalar getRelErr(const LTSResult& r) const;

										// get absolute error wrt r
	Scalar getAbsErr(const LTSResult& r) const;
};

//----------------------------------------------------------------------
//	SrchPriorList - the sequence of priorities of a node in the queue
//
//	A priority is a sequence of SrchPriority's. Values are compared
//	in various ways (see ALTSprQueue for further information).
//----------------------------------------------------------------------

class alts::SrchPriorList : public std::vector<SrchPriority> {
public:
										// constructor
	SrchPriorList() : std::vector<SrchPriority>() { }

										// construct from 1 arg
	SrchPriorList(SrchPriority a) : std::vector<SrchPriority>()
	{ push_back(a); }
										// construct from 2 args
	SrchPriorList(SrchPriority a1, SrchPriority a2) : std::vector<SrchPriority>()
	{ push_back(a1); push_back(a2); }

	void add(SrchPriority a)			// add a priority
	{ push_back(a); }

										// compare priorities
	bool operator<(const SrchPriorList& p) const;
};

//----------------------------------------------------------------------
//	alts::SrchParams
//	Used to store parameters related to the internals of the search
//	algorithm. See ALTS.cpp for default values.
//
//	The initSlopeRange is the range given by the user (in the
//	B*B:slope-range directive). The rooSlopeRange is the actual slope
//	range used by the root cell, which will either be this value, or the
//	result of slope sampling process. It's value is set in the search
//	algorithm.
//----------------------------------------------------------------------

class alts::SrchParams {
public:
										// general parameters
	Algorithm		algorithm;			// which search algorithm to use
	int				numPivots;			// number of improvement pivots

										// parameters for B&B
	SrchPriorList	srchPriorList;		// search priority list
	alts::PQOrder	pqOrder;			// prior queue order
	int				maxCells;			// max cells to expand
	int				cellsPerLevel;		// number of cells per level
	SlopeRange		initSlopeRange;		// initial slope range
	SlopeRange		rootSlopeRange;		// root cell slope range
	int				slopeSampleSize;	// sample size for initial cell
	Scalar			slopeExpansion;		// expansion factor for init cell
	bool			enableExpansion;	// allow range to be expanded
	Scalar			slopeTrimQuant;		// trimming factor for samples
	SplitMethod		splitMethod;		// cell splitting method

										// parameters for Fast-LTS
	int				numElemFits;		// number of elementary fits
	int				numBest;			// num of best fits to save

										// parameters for full-search-LTS
	int				fullSampleSize;		// number of samples
	int				fullMaxSlabs;		// max slabs to try

public:
	void setDefaults(int d);			// set default values

										// constructor from dim
	SrchParams(int d) : initSlopeRange(d), rootSlopeRange(d)
		{ setDefaults(d); }	
};

//----------------------------------------------------------------------
//	alts::Params
//	Used to store parameters related to the general execution of the
//	search algorithm. See ALTS.cpp for default values.
//----------------------------------------------------------------------

class alts::Params {
public:
	OutputLev		outputLev;			// output level
	std::ostream*	out;				// output stream
	int				dim;				// dimension of the space
	double			residAbsErr;		// residual absolute error
	double			residRelErr;		// residual related error
	int				trimAbsValue;		// absolute trimming parameter
	double			trimRelValue;		// relative trimming parameter
	int				trimAbsErr;			// absolute trimming error
	double			trimRelErr;			// relative trimming error
	int				trimLowerValue;		// lower trim value (absVal-absErr)
	int				traceFreq;			// trace frequency
	bool			doValidation;		// perform validation
	Plane			estimatedTrueFit;	// estimated true plane of fit
	PointSet		points;				// the points
	SrchParams		srchParams;			// search parameters

public:
	void setDefaults(int d);			// set defaults from dim

										// constructor from dim
	Params(int d = 1)
		: estimatedTrueFit(d), srchParams(d) { setDefaults(d); }
										// copy constructor (never used)
	Params(const Params& p)
		: estimatedTrueFit(0), srchParams(0) { assert(false); }

	void finalize();					// finalize values
};

//----------------------------------------------------------------------
//	SubsetEnum - subset enumerator
//
//	Enumerates all the subsets over a given integer range [0..n-1].
//
//	The subset is represented as a boolean vector. The method isTrivial
//	returns true if the set is empty or the complete set.
//
//	The reset() and next() return true if the current set contains a
//	valid subset.
//----------------------------------------------------------------------

class alts::SubsetEnum {
private:
	int					n;				// range
	std::vector<bool>	subset;			// the subset
	bool				isTrivialFlag;	// set is trivial
public:
	SubsetEnum(int nn);					// constructor

	int getSize() const					// get domain size
	{ return n; }
										// get subset
	const std::vector<bool>& getSubset() const
	{ return subset; }
										// index operator
	bool operator[](int i) const
	{ return subset[i]; }

	bool reset();						// reset to start

	bool next();						// next subset in the sequence

	bool isTrivial() const
	{ return isTrivialFlag; }
};

//----------------------------------------------------------------------
//	MSubsetEnum - m-element subset enumerator of {0,1,...,n-1}
//
//	The reset method sets the subset to its initial state {0,1,...,n-1}
//	and each call to next() advances the subset to the next subset in
//	the sequence lexicographically. When the subset value is the last
//	possible, that is, {n-m, n-m+1, ..., n-1}, the value false is
//	returned.
//
//	The reset() and next() return true if the current set contains a
//	valid subset.
//----------------------------------------------------------------------

class alts::MSubsetEnum {
public:
	int					n;				// range
	int					m;				// number of items
	std::vector<int>	subset;			// the subset
//
	MSubsetEnum(int nn, int mm);		// constructor

	int getDomainSize() const			// get domain size
	{ return n; }

	int getSetSize() const				// get set size
	{ return m; }
										// index operator
	int operator[](int i) const
	{ return subset[i]; }

	bool reset();						// reset to start

	bool next();						// next subset in the sequence

	bool has(int p) const;				// does current set contain p?
};

//----------------------------------------------------------------------
//	Output methods
//----------------------------------------------------------------------
										// output DVector
std::ostream& operator<<(std::ostream& out, const alts::DVector& x);
										// output Indep
std::ostream& operator<<(std::ostream& out, const alts::Indep& x);
										// output Plane
std::ostream& operator<<(std::ostream& out, const alts::Plane& b);
										// output Slope
std::ostream& operator<<(std::ostream& out, const alts::Slope& b);
										// output Point
std::ostream& operator<<(std::ostream& out, const alts::Point& p);
										// output LTSResult
std::ostream& operator<<(std::ostream& out, const alts::LTSResult& r);
										// output MSubsetEnum
std::ostream& operator<<(std::ostream& out, const alts::MSubsetEnum& s);
										// output SubsetEnum
std::ostream& operator<<(std::ostream& out, const alts::SubsetEnum& s);

//----------------------------------------------------------------------
//	Input methods
//----------------------------------------------------------------------
										// input Indep
std::istream& operator>>(std::istream& in, alts::Indep& x);
										// input Plane
std::istream& operator>>(std::istream& in, alts::Plane& b);
										// input Slope
std::istream& operator>>(std::istream& in, alts::Slope& b);
										// input Point
std::istream& operator>>(std::istream& in, alts::Point& p);

//----------------------------------------------------------------------
//	Comparison methods
//----------------------------------------------------------------------
										// compare DVector
bool operator==(const alts::DVector& x, const alts::DVector& y);
										// compare DVector
bool operator<=(const alts::DVector& x, const alts::DVector& y);

//----------------------------------------------------------------------
//	Additional utilities
//----------------------------------------------------------------------

namespace alts {

	void error(							// error routine
		const std::string&	msg,		// error message
		bool				abort);		// abort afterwards?

	void wait();						// wait for user input before exiting
										// used for Windows MS_Win32 version
	
										// affine combination of slopes
	Slope affineComb(Scalar a, const Slope& b1, const Slope& b2);
}

#endif
