//----------------------------------------------------------------------
// File:			ALTSfullSearchLTS.cpp
// Programmer:		David Mount
// Description:		Search using the Full-Search LTS heuristic
// Last modified:	06/17/2015 (Version 1.0)
//----------------------------------------------------------------------
// Copyright (c) 2015 University of Maryland and David Mount.  All
// Rights Reserved.
// 
// This software and related documentation is part of the Approximate
// Least-Trimmed Squares (ALTS) software.  This software is provided
// under the provisions of the GNU Public License (GPL). See the file
// ReadMe.txt for further information.
// 
// The University of Maryland (U.M.) and the authors make no
// representations about the suitability or fitness of this software for
// any purpose.  It is provided "as is" without express or implied
// warranty.
//----------------------------------------------------------------------
// History:
//	Revision 1.0  06/17/2015
//		Initial release
//----------------------------------------------------------------------

#include "ALTSfullSearchLTS.h"			// full-search LTS definitions
#include "ALTSnumeric.h"				// numeric utility definitions
#include "ALTSrand.h"					// random number generation

using namespace std;					// make std:: accessible
using namespace alts;					// make alts:: accessible

//----------------------------------------------------------------------
//	full-search LTS
//
//	This is an exhaustive search algorithm. We generate a random sample
//	of points (which might be the entire set). We enumerate all subsets
//	of size d+1. Each subset defines 2^{d+1} - 2 different slabs,
//	depending on which points lie on the top and bottom of the slab (at
//	least one point must be on each side). We invoke slabLTS to solve
//	the ordinary least squares for the points within the slab, and
//	return the best of all these.
//----------------------------------------------------------------------

LTSResult alts::fullSearchLTS(
	const Params&		params)			// parameters
{
	ostream* out = params.out;			// output stream
	const PointSet& pts = params.points;// the points
										// search parameters
	const SrchParams& srchParams = params.srchParams;
	int nPts = (int) pts.size();		// number of points
	int d = params.dim;					// dimension
	int m = d + 1;						// number of points in slab
	int h = params.trimAbsValue;		// trimming parameter
										// max slaps to try
	int maxSlabs = srchParams.fullMaxSlabs;

										// number of samples (0 => all)
	int nSamples = srchParams.fullSampleSize;
	PointSet samples;					// storage for samples

	LTSResult best(d);					// storage for the best result
	best.setVariance(SCALAR_INFTY);

	if (nSamples == 0) {				// use all points as samples
		PointSetConstIter it = pts.begin();
		while (it != pts.end()) {
			samples.push_back(*it);
			it++;
		}
		nSamples = nPts;				// actual number of samples
		if (srchParams.maxCells == 0 && srchParams.numPivots > 0) {
			error("Performing pivots with full enumeration is wasteful", false);
		}
	}
	else {								// generate random sample
		vector<int> sampIdx = randomSubset(nSamples, nPts);
		vector<int>::const_iterator it = sampIdx.begin();
		while (it != sampIdx.end()) {
			samples.push_back(pts[*it]);
			it++;
		}
	}
	if (nSamples < m) {					// too few samples to continue
		error("Too few samples for full-search LTS", true);
	}

	MSubsetEnum slabIdxSet(nSamples, m);// enum slab set indices
	bool moreSlabs = slabIdxSet.reset();
	int slabCount = 0;					// number of slabs tried
	while (moreSlabs && (maxSlabs == 0 ||  slabCount < maxSlabs )) {
										// tracing?
		bool doTrace = (params.traceFreq != 0
					&& slabCount % params.traceFreq == 0);
		if (doTrace) {
			*out << "Slab: " << slabCount
				 << " Index: " << slabIdxSet << "\n";
		}
		PointSet slabSet;				// generate slab set
		for (int i = 0; i < m; i++) {
			slabSet.push_back(pts[slabIdxSet[i]]);
		}
		SubsetEnum topSet(m);			// enumerator for top sets
		bool moreTopSets = topSet.reset();
		while (moreTopSets) {

			if(!topSet.isTrivial()){	// ignore trivial sets
				Plane bTop(d), bBot(d);	// slab planes
				bool success = 			// fit slab to slab set
					fitSlab(slabSet, topSet.getSubset(), bTop, bBot);

				if (success) {			// good slab -- get LTS cost
					LTSResult r = slabLTS(pts, slabIdxSet, h, bTop, bBot); 
					r.setTimeStamp(slabCount);
										// successful?
					if (r.getVariance() != SCALAR_INFTY) {
						if (doTrace) {
							*out << "  TopSet: " << topSet
								 << " result: " << r << "\n";
						}
										// apply improvement steps
						for (int j = 0; j < srchParams.numPivots; j++) {
							bool improved = r.improveFit(pts, h);
							if (doTrace) {
								*out << "    Improved:" << j
									 << " result: " << r << "\n";
							}
							if (!improved) break;
						}
										// update if new best
						if (r.getVariance() < best.getVariance()) {
							best = r;
							if (doTrace) *out << "  [best]\n";
						}
					}
				}
			}
			moreTopSets = topSet.next(); // get next top set
		}
		moreSlabs = slabIdxSet.next();	// get next slab set
		slabCount++;
	}
	return best;						// return best result
}
