//----------------------------------------------------------------------
// File:			ALTSnumeric.cpp
// Programmer:		David Mount
// Description:		Numerical utilities
// Last modified:	06/17/2015 (Version 1.0)
//----------------------------------------------------------------------
// Copyright (c) 2015 University of Maryland and David Mount.  All
// Rights Reserved.
// 
// This software and related documentation is part of the Approximate
// Least-Trimmed Squares (ALTS) software.  This software is provided
// under the provisions of the GNU Public License (GPL). See the file
// ReadMe.txt for further information.
// 
// The University of Maryland (U.M.) and the authors make no
// representations about the suitability or fitness of this software for
// any purpose.  It is provided "as is" without express or implied
// warranty.
//----------------------------------------------------------------------
// History:
//	Revision 1.0  06/17/2015
//		Initial release
//----------------------------------------------------------------------

#include "ALTSnumeric.h"				// numerical definitions
#include "ALTSbasicStat.h"				// basic stat definitions

#include <algorithm>					// for max()

using namespace std;					// make std:: accessible
using namespace alts;					// make alts:: accessible

//----------------------------------------------------------------------
//	Local Utilities
//----------------------------------------------------------------------
										// safe sqrt(a^2 + b^2)
static Scalar pythag(
	Scalar a,
	Scalar b)
{
	Scalar at = fabs(a);
	Scalar bt = fabs(b);
	if (at > bt) {
		Scalar ct = bt/at;
		return at * sqrt(1.0 + ct*ct);
	}
	else if (bt != 0) {
		Scalar ct = at/bt;
		return bt * sqrt(1.0 + ct*ct);
	}
	else {
		return 0;
	}
}

static Scalar sign(						// |a| * sign(b)
	Scalar a,
	Scalar b)
{
	return (b >= 0.0 ? fabs(a) : -fabs(a));
}

//----------------------------------------------------------------------
//	Vector/Matrix allocation/deallocation
//
//	To support Fortran-style indexing, the pointers are offset so that
//	v[1] is the first element.
//----------------------------------------------------------------------

static Scalar* allocVector(				// allocate vector[nl..nh]
	int			nl,
	int			nh)
{
	Scalar* v= new Scalar[nh - nl + 1];
	return v - nl;
}

static void freeVector(					// free vector
	Scalar*		v,
	int			nl,
	int			nh)
{
	delete [] (v + nl);
}

static Scalar** allocMatrix(			// allocate matrix
	int			nrl,
	int			nrh,
	int			ncl,
	int			nch)
{
	typedef Scalar* ScalarPtr;			// row pointers
	Scalar** m = new ScalarPtr[nrh - nrl + 1];
	m -= nrl;

	for(int i = nrl; i <= nrh; i++) {	// column storage
		m[i] = new Scalar[nch - ncl + 1];
		m[i] -= ncl;
	}
	return m;
}

static void freeMatrix(					// free matrix
	Scalar**	m,
	int			nrl,
	int			nrh,
	int			ncl,
	int			nch)
{
	for(int i = nrh; i >= nrl; i--) delete [] (m[i] + ncl);
	delete [] (m + nrl);
}

//----------------------------------------------------------------------
//	IntResidSq - an integer, a residual, and a squared residual
//----------------------------------------------------------------------

class IntResidSq {
public:
	int		id;							// integer identifier
	Scalar	res;						// residual
	Scalar	sq;							// squared residual
public:
										// constructor from parts
	IntResidSq(int ii = 0, Scalar rr = 0, Scalar ss = 0)
		: id(ii), res(rr), sq(ss) { }
};

bool operator<(const IntResidSq& irs1, const IntResidSq& irs2)
{  return irs1.sq < irs2.sq; }

//----------------------------------------------------------------------
//	svdcmp - singular value decomposition
//
//	Given a matrix a[1..m][1..n] this routine computes it singular value
//	decomposition, A = U*W*V^T.  The matrix U replaces a on output. The
//	diagonal matrix of singular values W is output as a vector w[1..n].
//	The matrix V (not the transpose) is output as v[1..n][1..n].
//
//	Returns false if svdcmp does not terminate in 30 iterations.
//
//	WARNING: This code was copied from somewhere, but I cannot remember.
//	It was probably "Numerical Recipes in C" by William Press, Brian
//	Flannery, Saul Teukolsky, and William Vetterling.
//----------------------------------------------------------------------

static bool svdcmp(						// singular value decomp
	Scalar**	a,						// A = U[1..m][1..n] (modified)
	int			m,
	int			n,
	Scalar*		w,						// W[1..n]
	Scalar**	v)						// V[1..n][1..n]
{
	int flag, i, its, j, jj, k;
	int l = 0;
	int nm = 0;
	Scalar c, f, h, s, x, y, z;
	Scalar anorm = 0.0, g = 0.0, scale = 0.0;
	Scalar *rv1;

	rv1 = allocVector(1, n); 
	for (i = 1; i <= n; i++) {
		l = i + 1;
		rv1[i] = scale*g;
		g = s = scale = 0.0;
		if (i <= m) {
			for (k = i; k <= m; k++) scale += fabs(a[k][i]);
			if (scale) {
				for (k = i; k <= m; k++) {
					a[k][i] /= scale;
					s += a[k][i]*a[k][i];
				}
				f = a[i][i];
				g = -sign(sqrt(s), f);
				h = f*g - s;
				a[i][i] = f - g;
				if (i != n) {
					for (j = l; j <= n; j++) {
						for (s = 0.0, k = i; k <= m; k++)
							s += a[k][i]*a[k][j];
						f = s/h;
						for (k = i; k <= m; k++)
							a[k][j] += f*a[k][i];
					}
				}
				for (k = i; k <= m; k++) a[k][i] *= scale;
			}
		}
		w[i] = scale*g;
		g = s = scale = 0.0;
		if (i <= m && i != n) {
			for (k = l; k <= n; k++) scale += fabs(a[i][k]);
			if (scale) {
				for (k = l; k <= n; k++) {
					a[i][k] /= scale;
					s += a[i][k]*a[i][k];
				}
				f = a[i][l];
				g = -sign(sqrt(s), f);
				h = f*g - s;
				a[i][l] = f - g;
				for (k = l; k <= n; k++) rv1[k] = a[i][k]/h;
				if (i != m) {
					for (j = l; j <= m; j++) {
						for (s = 0.0, k = l; k <= n; k++)
							s += a[j][k]*a[i][k];
						for (k = l; k <= n; k++)
							a[j][k] += s*rv1[k];
					}
				}
				for (k = l; k <= n; k++) a[i][k] *= scale;
			}
		}
		anorm = max(anorm, (fabs(w[i]) + fabs(rv1[i])));
	}
	for (i = n; i >= 1; i--) {
		if (i < n) {
			if (g) {
				for (j = l; j <= n; j++)
					v[j][i] = (a[i][j]/a[i][l])/g;
				for (j = l; j <= n; j++) {
					for (s = 0.0, k = l; k <= n; k++)
						s += a[i][k]*v[k][j];
					for (k = l; k <= n; k++)
						v[k][j] += s*v[k][i];
				}
			}
			for (j = l; j <= n; j++) v[i][j] = v[j][i] = 0.0;
		}
		v[i][i] = 1.0;
		g = rv1[i];
		l = i;
	}
	for (i = n; i >= 1; i--) {
		l = i + 1;
		g = w[i];
		if (i < n)
			for (j = l; j <= n; j++) a[i][j] = 0.0;
		if (g) {
			g = 1.0/g;
			if (i != n) {
				for (j = l; j <= n; j++) {
					for (s = 0.0, k = l; k <= m; k++)
						s += a[k][i]*a[k][j];
					f = (s/a[i][i])*g;
					for (k = i; k <= m; k++)
						a[k][j] += f*a[k][i];
				}
			}
			for (j = i; j <= m; j++) a[j][i] *= g;
		} else {
			for (j = i; j <= m; j++) a[j][i] = 0.0;
		}
		++a[i][i];
	}
	for (k = n; k >= 1; k--) {
		for (its = 1; its <= 30; its++) {
			flag = 1;
			for (l = k; l >= 1; l--) {
				nm = l - 1;
				if (fabs(rv1[l]) + anorm == anorm) {
					flag = 0;
					break;
				}
				if (fabs(w[nm]) + anorm == anorm) break;
			}
			if (flag) {
				c = 0.0;
				s = 1.0;
				for (i = l; i <= k; i++) {
					f = s*rv1[i];
					if (fabs(f) + anorm != anorm) {
						g = w[i];
						h = pythag(f, g);
						w[i] = h;
						h = 1.0/h;
						c = g*h;
						s = (-f*h);
						for (j = 1; j <= m; j++) {
							y = a[j][nm];
							z = a[j][i];
							a[j][nm] = y*c + z*s;
							a[j][i] = z*c - y*s;
						}
					}
				}
			}
			z = w[k];
			if (l == k) {
				if (z < 0.0) {
					w[k] = -z;
					for (j = 1; j <= n; j++) v[j][k] = (-v[j][k]);
				}
				break;
			}
			if (its == 30) {			// no convergence
				freeVector(rv1, 1, n);
				return false;
			}
			x = w[l];
			nm = k - 1;
			y = w[nm];
			g = rv1[nm];
			h = rv1[k];
			f = ((y - z)*(y + z) + (g - h)*(g + h))/(2.0*h*y);
			g = pythag(f, 1.0);
			f = ((x - z)*(x + z) + h*((y/(f + sign(g, f))) - h))/x;
			c = s = 1.0;
			for (j = l; j <= nm; j++) {
				i = j + 1;
				g = rv1[i];
				y = w[i];
				h = s*g;
				g = c*g;
				z = pythag(f, h);
				rv1[j] = z;
				c = f/z;
				s = h/z;
				f = x*c + g*s;
				g = g*c - x*s;
				h = y*s;
				y = y*c;
				for (jj = 1; jj <= n; jj++) {
					x = v[jj][j];
					z = v[jj][i];
					v[jj][j] = x*c + z*s;
					v[jj][i] = z*c - x*s;
				}
				z = pythag(f, h);
				w[j] = z;
				if (z) {
					z = 1.0/z;
					c = f*z;
					s = h*z;
				}
				f = (c*g) + (s*y);
				x = (c*y) - (s*g);
				for (jj = 1; jj <= m; jj++) {
					y = a[jj][j];
					z = a[jj][i];
					a[jj][j] = y*c + z*s;
					a[jj][i] = z*c - y*s;
				}
			}
			rv1[l] = 0.0;
			rv1[k] = f;
			w[k] = x;
		}
	}
	freeVector(rv1, 1, n);
	return true;
}

//----------------------------------------------------------------------
//	svbksb - Singular value back-substitution
//
//	Solves A*X = B for a vector X, where A is specified by the arrays
//	u[1..m][1..n], w[1..n], v[1..n][1..n] as returned by svdcmp().  m
//	and n are the dimensions of a, and will be be equal for square
//	matrices.  b[1..m] is the input right-hand side.  x[1..n] is the
//	output solution vector.  No input quantities are destroyed, so the
//	routine may be called sequentially with different b's.
//
//	Returns false if something goes wrong in SVD.
//
//	WARNING: This code was copied from somewhere, but I cannot remember.
//	It was probably "Numerical Recipes in C" by William Press, Brian
//	Flannery, Saul Teukolsky, and William Vetterling.
//----------------------------------------------------------------------

static void svbksb(
	Scalar**	u,						// U[1..m][1..n]
	Scalar*		w,						// W[1..n]
	Scalar**	v,						// V[1..n][1..n]
	int			m,
	int			n,
	Scalar*		b,						// B[1..m]
	Scalar*		x)						// X[1..n] (returned)
{
	int jj, j, i;
	Scalar s, *tmp;

	tmp = allocVector(1, n);
	for (j = 1; j <= n; j++) {
		s = 0.0;
		if (w[j]) {
			for (i = 1; i <= m; i++) s += u[i][j]*b[i];
			s /= w[j];
		}
		tmp[j] = s;
	}
	for (j = 1; j <= n; j++) {
		s = 0.0;
		for (jj = 1; jj <= n; jj++) s += v[j][jj]*tmp[jj];
		x[j] = s;
	}
	freeVector(tmp, 1, n);
}

//----------------------------------------------------------------------
//	svdfit - Use singular value decomposition for least squares fit
//
//	This is a modification of the procedure svdfit from Numerical
//	Recipes in C for the purpose of fitting Given a set of ndata data
//	points x[1..ndata][1..ma] in ma dimensional space and ndata
//	observations, y[1..ndata], use least squares minimization to
//	determine the coefficients a[1..ma] of the fitting function
//
//		y_j = SUM_i a_i * x_{j,i} + err_j, for j = 1..ndata, i = 1..ma
//
//	The fitting equation is solved using singular value decomposition of
//	the (ndata x ma) matrix x (which takes the role of u in the SVD. The
//	array x is destroyed in the process, and arrays v[1..ma][1..ma] and
//	w[1..ma] provide workspace on input. On output, x, v, and w define
//	the singular value decomposition, and can be used to obtain the
//	covariance matrix. The program returns values for the ma fit
//	parameters a, and the sum of squared residuals in chiSq.
//----------------------------------------------------------------------

static bool svdfit(					// solve X*A = Y via SVD
	Scalar**	x,					// X[1..nd][1..ma]
	Scalar*		y,					// Y[1..nd]
	int			ndata,				// nd (must be >= ma)
	Scalar*		a,					// A[1..ma] (returned)
	int			ma,
	Scalar**	u,					// U[1..nd][1..ma] (returned)
	Scalar**	v,					// V[1..ma][1..ma] (returned)
	Scalar*		w,					// W[1..ma] (returned)
	Scalar&		chiSq)				// squared error (returned)
{
	const Scalar TOL = 1.0e-5;		// floating point tolerance

	int j, i;
	Scalar wmax, tmp, thresh, sum;

	for (i = 1; i <= ndata; i++) {
		for (j = 1; j <= ma; j++) u[i][j] = x[i][j];
	}
	bool okay = svdcmp(u, ndata, ma, w, v);
	if (!okay) return false;
	wmax = 0.0;
	for (j = 1; j <= ma; j++)
		if (w[j] > wmax) wmax = w[j];
	thresh = TOL*wmax;
	for (j = 1; j <= ma; j++)
		if (w[j] < thresh) w[j] = 0.0;
	svbksb(u, w, v, ndata, ma, y, a);
	chiSq = 0.0;
	for (i = 1; i <= ndata; i++) {
		for (sum = 0.0, j = 1; j <= ma; j++) sum += a[j]*x[i][j];
		chiSq += (tmp = (y[i] - sum), tmp*tmp);
	}
	return true;
}

//----------------------------------------------------------------------
// olsFit - compute an ordinary least squares fit
//
//	Computes the plane of fit for a PointSet pts of m >= d points
//	using ordinary least squares. Each point consists of independent
//	variables (1, x1, x2, ..., xd) and dependent variable y, and the
//	result is a Plane object b = (b0, b1, ..., b{d-1}) so that
//
//		y = b0 + b1*x1 + b2*x2 + ... + b{d-1}*x{d-1} + error
//
//	We do this by invoking the procedure alts::svdfit(). Note that this
//	procedure assumes that vectors and matrices are indexed starting at
//	1, thus, we need to do some index translation when copying elements
//	from our arrays into the ones used by svdfit().
//
//	The system is defined by the matrix system
//
//		X * a = y,
//
//	where X[1..m][1..d] is a matrix whose rows are the independent
//	variables (we exploit the fact that x0 = 1 in our internal
//	representation of independent vectors), the vector y[1..m], whose
//	elements are the corresponding dependent variables, and the solution
//	vector is a[1..d], which is mapped to b[0..d-1].
//
//	This procedure can be used either for computing an elementary fit
//	(when the number of points equals the dimension) or performing least
//	squares fit (when the number of points exceeds the dimension).
//
//	Note that the coefficient of the plane b are all modified on return.
//	The plane b needs to be defined on entry, because we use it to
//	determine the dimension of the space.
//
//	Returns false if something goes wrong numerically.
//----------------------------------------------------------------------

bool alts::olsFit(
	const PointSet&		pts,			// the points
	Plane&				b)				// plane of fit (modified)
{
	int nRows = (int) pts.size();		// number of points
	int nCols = b.getDim();				// dimension

										// allocate temp storage
	Scalar** X = allocMatrix(1, nRows, 1, nCols);
	Scalar*  y = allocVector(1, nRows);
	Scalar*  a = allocVector(1, nCols);

	Scalar** U = allocMatrix(1, nRows, 1, nCols);
	Scalar** V = allocMatrix(1, nCols, 1, nCols);
	Scalar*  w = allocVector(1, nCols);
	Scalar chiSq;						// holds residual error

	pts.shiftIndepTo(X, 1, 1);			// X[1..][1..] = indep[0..][0..]
	pts.shiftDepTo(y, 1);				// Y[1..] = dep[0..]
										// perform the fit
	bool isOkay = svdfit(X, y, nRows, a, nCols, U, V, w, chiSq);

	b.shiftCopyFrom(a, 1);				// b[0..] = a[1..]

	freeVector(w, 1, nCols);			// free temp storage
	freeMatrix(V, 1, nCols, 1, nCols);
	freeMatrix(U, 1, nRows, 1, nCols);
	freeVector(a, 1, nCols);
	freeVector(y, 1, nRows);
	freeMatrix(X, 1, nRows, 1, nCols);

	return isOkay;						// return status
}

//----------------------------------------------------------------------
// fitSlab - fits a slab to a collection of d+1 points
//
//	A slab is a pair of parallel planes. This procedure computes a slab
//	defined by a set of d+2 points and a specification of which points
//	are on top and which are on the bottom (through the boolean vector
//	onTop[]). Each point consists of independent variables (1, x1, x2,
//	..., xd) and dependent variable y, and the result is a Plane objects
//	bTop = (bt0, b1, ..., b{d-1}) and bBot = (bb0, b1, ..., b{d-1})
//	such that
//
//		y = bt0 + b1*x1 + ... + b{d-1}*x{d-1}   for all (x,y) on top
//
//		y = bb0 + b1*x1 + ... + b{d-1}*x{d-1}   for all (x,y) on bottom
//
//	We do this by invoking the procedure alts::svdfit() on the following
//	example (d+1)x(d+1) system X * a = y:
//
//		[ 1 0 x11 x12 ... x1{d-1} ] [ bt0 ]  =  [ y1 ] : y1 on top
//		[ 0 1 x21 x22 ... x2{d-1} ] [ bb0 ]  =  [ y2 ] : y2 on bottom
//		[ 1 0 x31 x32 ... x3{d-1} ] [  b1 ]  =  [ y3 ] : y3 on top
//		[ 1 0 x41 x42 ... x4{d-1} ] [  b2 ]  =  [ y4 ] : y4 on top
//		[ 0 1 x51 x52 ... x5{d-1} ] [  b3 ]  =  [ y5 ] : y5 on bottom
//
//	This procedure assumes that vectors and matrices are indexed
//	starting at 1, thus, we need to do some index translation when
//	copying elements from our arrays into the ones used by svdfit().
//
//	From the result vector a we form the vectors (bt0, b1, ..., b{d-1})
//	and (bb0, b1, ..., b{d-1}), for the top and bottom planes,
//	respectively.
//
//	Note that the coefficient of the plane b are all modified on return.
//	The plane b needs to be defined on entry, because we use it to
//	determine the dimension of the space.
//
//	Returns false if something goes wrong numerically.
//----------------------------------------------------------------------

bool alts::fitSlab(
	const PointSet&		pts,			// the points
	vector<bool>		onTop,			// which points are on top
	Plane&				bTop,			// slab's top plane (modified)
	Plane&				bBot)			// slab's bottom plane (modified)
{
	int d = bTop.getDim();				// dimension of the space
										// check consistency
	assert(bBot.getDim() == d && int(pts.size()) == d+1);
	int m = d + 1;						// dimension of system

										// allocate temp storage
	Scalar** X = allocMatrix(1, m, 1, m);
	Scalar*  y = allocVector(1, m);
	Scalar*  a = allocVector(1, m);

	Scalar** U = allocMatrix(1, m, 1, m);
	Scalar** V = allocMatrix(1, m, 1, m);
	Scalar*  w = allocVector(1, m);
	Scalar chiSq;						// holds residual error

	for (int i = 0; i < m; i++) {
		const Indep& xi = pts[i].getIndep();
		xi.shiftCopyTo(X[i+1], 2);		// x[0..d-1] to X[i+1][2..m]
		if (onTop[i]) {					// set X[i+1][0] and [1]
			X[i+1][1] = 1; X[i+1][2] = 0;
		}
		else {
			X[i+1][1] = 0; X[i+1][2] = 1;
		}
		y[i+1] = pts[i].getDep();		// dep[i] to y[i+1]
	}

										// solve the system
	bool isOkay = svdfit(X, y, m, a, m, U, V, w, chiSq);

	bTop[0] = a[1];						// store result coefficients
	bBot[0] = a[2];
	for (int i = 1; i < d; i++) {
		bTop[i] = bBot[i] = a[i+2];
	}

	freeVector(w, 1, m);				// free temp storage
	freeMatrix(V, 1, m, 1, m);
	freeMatrix(U, 1, m, 1, m);
	freeVector(a, 1, m);
	freeVector(y, 1, m);
	freeMatrix(X, 1, m, 1, m);

	return isOkay;						// return status
}

//----------------------------------------------------------------------
//	LTSResult::improveFit - locally improve an LTS fit
//
//	Local improvement is given a point set, trimming value h.  It
//	computes the h closest points to the current plane and computes and
//	returns their ordinary least square plane of fit. We update the
//	LTSResult. The result time-stamp remains unchanged.
//
//	The procedure returns false if the result plane is the same as the
//	original.
//----------------------------------------------------------------------

bool LTSResult::improveFit(
	const PointSet&		pts,			// the points
	int					h)				// trimming parameter
{
	int n = (int) pts.size();			// number of points
	int d = b.getDim();					// dimension
	vector<IntResidSq> resids;			// storage for indices and residuals
	resids.reserve(n);					// space for n residuals

	for (int i = 0; i < n; i++) {		// store id's and residuals
		Scalar res = b.residual(pts[i]);
		Scalar sq = res*res;
		resids.push_back(IntResidSq(i, res, sq));
	}
										// select the h-th smallest
	nth_element(resids.begin(), resids.begin()+(h-1), resids.end());
	

	PointSet inliers;					// storage for inliers
	for (int i = 0; i < h; i++) {		// insert inliers into set
		inliers.insert(pts[resids[i].id]);
	}

	Plane b2(d);						// updated plane
	bool success = olsFit(inliers, b2);	// attempt OLS fit to inliers
	if (!success) {
		error("Internal error: Cannot find fit in improveLTSfit", true);
	}
	if (b2 == b) return false;			// no change --> converged

	BasicStat stat;						// compute variance
	for (int i = 0; i < h; i++) {
		stat += b2.residual(inliers[i]);
	}
	Scalar thisVariance = stat.variance();
	if (thisVariance >= variance) {		// no improvement?
		return false;
	}
	variance = thisVariance;			// update variance
	b = b2;								// update plane
	return true;
}

//----------------------------------------------------------------------
//	getLTSvariance - compute LTS variance of plane
//
//	Computes the variance of the residuals of a given point set with
//	respect to a given plane and a given number of inliers.
//----------------------------------------------------------------------

Scalar alts::getLTSvariance(
	const PointSet&		pts,			// the points
	int					h,				// trimming parameter
	const Plane&		b)				// initial fitting plane
{
	int n = (int) pts.size();			// number of points
	vector<IntResidSq> resids;			// storage for indices and residuals
	resids.reserve(n);					// space for n residuals

	for (int i = 0; i < n; i++) {		// store id's and residuals
		Scalar res = b.residual(pts[i]);
		Scalar sq = res*res;
		resids.push_back(IntResidSq(i, res, sq));
	}
										// select the h-th smallest
	nth_element(resids.begin(), resids.begin()+(h-1), resids.end());

	BasicStat stat;						// collect residuals of the inliers
	for (int i = 0; i < h; i++) {
		stat += resids[i].res;
	}

	return stat.variance();				// return the variance
}

//----------------------------------------------------------------------
//	slabLTS - compute LTS solution of points in a slab
//
//	We are given a point set, the indices of the points defining a slab,
//	(called the slab-set of size m) the desired number of inliers (h),
//	and the slab bounding planes. Let I denote the subset of points that
//	lie within the slab, excluding the points defining the slab. Let nIn
//	denote the size of I. If n exceeds h or is less than h-m then this
//	slab is infeasible (since it cannot contain h points) and we return
//	SCALAR_INFTY. We consider all possible subsets S of h-nIn elements
//	of the slabset, and for each we compute the ordinary least squares
//	variance of the h-element set (I union S). We return the variance of
//	smallest variance.
//
//	TODO: There is a bug if the points are not in general position and
//	points other than the slab points lie on the slab.
//----------------------------------------------------------------------

LTSResult alts::slabLTS(
	const PointSet&		pts,			// the point set
	const MSubsetEnum&	slabIdxSet,		// the points defining the slab
	int					h,				// trimming parameter
	const Plane&		bTop,			// top of slab
	const Plane&		bBot)			// bottom of slab
{
	int d = (int) bTop.size();			// dimension
	int n = (int) pts.size();			// number of points
	int m = slabIdxSet.getSetSize();	// number in slab set
	int nInMin = h - m;					// min number of inliers
	int nInMax = h;						// max number of inliers
	PointSet inliers;					// storage for points in slab

	LTSResult best(d);					// storage for best result
	best.setVariance(SCALAR_INFTY);

	int nIn = 0;						// number of inliers

	for (int i = 0; i < n; i++) {		// find points in slab
										// residuals w.r.t. slab
		Scalar resTop = bTop.residual(pts[i]);
		Scalar resBot = bBot.residual(pts[i]);
										// point lies within slab?
		if (resTop < 0 && resBot > 0 && !slabIdxSet.has(i)) {
			inliers.push_back(pts[i]);	// add to inliers
			nIn++;
		}
		if (nIn > nInMax) return best;	// found too many inliers?
	}

	if (nIn < nInMin) return best;		// found too few inliers

	int nExtras = h - nIn;				// number of slab points needed
	assert(nExtras >= 0 && nExtras <= m);

	MSubsetEnum extras(m, nExtras);
	bool moreExtras = extras.reset();
	while (moreExtras) {				// consider all slab pts to include
										// add slab points to inliers
		for (int i = 0; i < nExtras; i++) {
			inliers.push_back(pts[slabIdxSet[extras[i]]]);
		}
		assert(int(inliers.size()) == h); // should have exactly h pts now
										// perform OLS on inliers
		Plane b(d);
		bool success = olsFit(inliers, b);
		if (!success) {
			error("Internal error: Cannot find fit in slabLTS", true);
		}
										// get its variance
		Scalar var = getLTSvariance(pts, h, b);
										// improves previous result?
		if (var < best.getVariance()) {
			best = LTSResult(var, b);
		}
										// remove slab points from inliers
		for (int i = 0; i < nExtras; i++) {
			inliers.pop_back();
		}
		moreExtras = extras.next();
	}
	return best;						// return best result
}
