//----------------------------------------------------------------------
// File:			ALTSprQueue.cpp
// Programmer:		David Mount
// Description:		ALTS priority queue
// Last modified:	06/17/2015 (Version 1.0)
//----------------------------------------------------------------------
// Copyright (c) 2015 University of Maryland and David Mount.  All
// Rights Reserved.
// 
// This software and related documentation is part of the Approximate
// Least-Trimmed Squares (ALTS) software.  This software is provided
// under the provisions of the GNU Public License (GPL). See the file
// ReadMe.txt for further information.
// 
// The University of Maryland (U.M.) and the authors make no
// representations about the suitability or fitness of this software for
// any purpose.  It is provided "as is" without express or implied
// warranty.
//----------------------------------------------------------------------
// History:
//	Revision 1.0  06/17/2015
//		Initial release
//----------------------------------------------------------------------

#include <algorithm>					// provides push_heap, pop_heap
#include "ALTSrand.h"					// random number generation
#include "ALTSprQueue.h"				// priority queue includes

using namespace std;					// make std:: accessible
using namespace alts;					// make alts:: accessible

//----------------------------------------------------------------------
//	See ALTSprQueue.h for general information. Note that the current key
//	is initialized to -1 and incremented before going to the next key in
//	cyclic order. This is done so that the index currKey reflects the
//	index of the most recently removed key. This is important for the
//	methods that do rewarding/punshing of key selection, since they are
//	applied to the key of the most recently extracted item.
//----------------------------------------------------------------------

CellPrQueue::CellPrQueue()				// default constructor
	: order(LEX_ORDER), nKeys(0), currKey(-1), vq(), lq() { }

bool CellPrQueue::usingLexOrder() const	// using lexicographic order?
	{ return order == LEX_ORDER; }

bool CellPrQueue::usingCyclicOrder() const	// using cyclic order?
	{ return order == CYCLIC_ORDER; }

bool CellPrQueue::usingWeightedOrder() const // using weighted order?
	{ return order == WEIGHTED_ORDER; }

void CellPrQueue::clear()				// clear everything
{
	vq.clear();
	lq.clear();
	weights.clear();
}

void CellPrQueue::useLexOrder()			// use lexicographic order
{
	order = LEX_ORDER;
	clear();
}

void CellPrQueue::useCyclicOrder(int nk)	// use cyclic order
{
	order = CYCLIC_ORDER;
	clear();
	nKeys = nk;
	currKey = -1;
}

void CellPrQueue::useWeightedOrder(int nk)	// use weighted order
{
	order = WEIGHTED_ORDER;
	clear();
	nKeys = nk;
	weights = vector<double>(nk, 1);	// assign uniform weights
	normalizeWeights();					// normalize them
}

bool CellPrQueue::isEmpty() const		// is queue empty?
	{ return (usingLexOrder() ? vq.empty() : lq.empty()); }

int CellPrQueue::size() const			// size of queue
	{ return (usingLexOrder() ? int(vq.size()) : int(lq.empty())); }

void CellPrQueue::insert(SlopeCell*	inf) // insert item into queue
{
	if (usingLexOrder()) {					// using lexicographic queue?
		vq.push_back(CellPqNode(inf->priorities, inf));
		push_heap(vq.begin(), vq.end());	// sift-up
	}
	else {
		lq.push_back(CellPqNode(inf->priorities, inf));
	}
}

void CellPrQueue::nextCyclicKey()		// advance to next cyclic key
{
	currKey = (currKey + 1) % nKeys;
}

// Beware: This actually extracts the max key, not min!

SlopeCell* CellPrQueue::extractMin()	// extract minimum
{
	if (usingLexOrder()) {				// lex ordering?
		SlopeCell* c = vq.front().info;	// front of queue
		pop_heap(vq.begin(), vq.end());	// push to back of queue
		vq.pop_back();					// remove it
		return c;
	}
	else {								// cyclic or weighted ordering?
		int useKey;						// key to use
		if (usingCyclicOrder()) {		// cyclic key order?
			nextCyclicKey();			// advance to next key
			useKey = currKey;			// use current cyclic key
		}
		else {							// weighted order?
			double r = ranUniform(0, 1);// random uniform
			useKey = 0;					// start at 0
			while (useKey < nKeys && r > weights[useKey]) {
				r -= weights[useKey];	// decrement r
				useKey++;				// try next key
			}
			currKey = useKey;			// this becomes the current key
		}
		LIterator p = lq.begin();		// start of queue
		LIterator maxNode = p;			// current max node
		while (++p != lq.end()) {
			if (maxNode->isLessThan(*p, useKey)) {
				maxNode = p;
			}
		}
		SlopeCell* c = maxNode->info;	// associated cell
		lq.erase(maxNode);				// remove it
		return c;						// return max cell
	}
}

										// get minimum lower bound
Scalar CellPrQueue::getMinLowerBound() const
{
	Scalar minLowerBnd = SCALAR_INFTY;	// minimum lower bound so far
	if (usingLexOrder()) {				// using lex order - vq
		ConstVIterator p = vq.begin();	// start of queue
		while (p != vq.end()) {
			if (!p->info->haveLower) alts::error("Expect lower bound", true);
										// update min
			minLowerBnd = min(minLowerBnd, p->info->lowerBnd);
			++p;						// advance
		}
	}
	else {								// using cyclic order - lq
		ConstLIterator p = lq.begin();	// start of queue
		while (p != lq.end()) {
			if (!p->info->haveLower) alts::error("Expect lower bound", true);
										// update min
			minLowerBnd = min(minLowerBnd, p->info->lowerBnd);
			++p;						// advance
		}
	}
	return minLowerBnd;
}

void CellPrQueue::normalizeWeights()	// normalize weights
{
	double sum = 0;
	for (int i = 0; i < nKeys; i++) {	// total weight
		sum += weights[i];
	}
	if (sum == 0) {
		error("Internal error: weights cannot all be zero", true);
	}
										// min and max weights
	double minWeight = 1/double(2*nKeys);
	double maxWeight = double(nKeys + 1)/double(2*nKeys);
	double sum2 = 0;
	for (int i = 0; i < nKeys; i++) {	// cap weights
		weights[i] /= sum;				// initial normalization
		if (weights[i] < minWeight) {
			weights[i] = minWeight;
		} else if (weights[i] > maxWeight) {
			weights[i] = maxWeight;
		}
		sum2 += weights[i];				// adjusted sum
	}
	for (int i = 0; i < nKeys; i++) {	// normalize adjusted weights
		weights[i] /= sum2;
	}
}
										// reward current weight
void CellPrQueue::rewardWeight(double score)
{
										// not weighted or not initialized
	if (!usingWeightedOrder() || currKey < 0) return;
	double r = ranUniform(0, 1);		// random uniform
	if (score >= r) {					// is score high enough?
		weights[currKey] *= 1.5;		// yes - increase weight
		normalizeWeights();				// renormalize
	}
	else {
		weights[currKey] *= 0.90;		// yes - decrease weight
		normalizeWeights();				// renormalize
	}
}

void CellPrQueue::printWeights(ostream* out) const	// print weight info
{
	if (!usingWeightedOrder()) return;	// not using weights
	*out << "[";
	*out << weights[0];
	for (int i = 1; i < nKeys; i++) {
		*out << "," << weights[i];
	}
	*out << "]: curr = " << currKey;
}

void CellPrQueue::print(ostream* out) const	// print contents
{
	const int MAX_TO_PRINT = 20;		// print at most this many entries
	*out << "Priority queue contents: ";
	if (usingLexOrder()) {				// using lex order - vq
		ConstVIterator p = vq.begin();	// start of queue
		int nPrinted = 0;				// number of entries printed
		*out << "(Lexicographic ordering)\n";
		while (p != vq.end()) {
			*out << "  [" << p->info->cellNum << "]: " << p->key;
			*out << " upper: ";
			if (p->info->haveUpper)	*out << p->info->upperBnd.getVariance();
			else *out << "----";
			*out << " lower: ";
			if (p->info->haveLower)	*out << p->info->lowerBnd;
			else *out << "----";
			*out << "\n";
			++p;						// advance
			if (++nPrinted >= MAX_TO_PRINT) {
				int remainder = vq.size() - MAX_TO_PRINT;
				if (remainder > 0) {
					*out << "  ...followed by " << vq.size() - MAX_TO_PRINT << " more...\n";
				}
				break;
			}
		}
	}
	else {								// cyclic or weighted order
		if (usingCyclicOrder()) {		// cyclic - print current key
			*out << "(Cyclic ordering: " << currKey << ")\n";
		}
		else {							// weighted - print weights
			*out << "(Weighted order: ";
			printWeights(out);
			*out << ")\n";
		}
		ConstLIterator p = lq.begin();	// start of queue
		int nPrinted = 0;				// number of entries printed
		while (p != lq.end()) {
			*out << "  [" << p->info->cellNum << "]: " << p->key;
			*out << " upper: ";
			if (p->info->haveUpper)	*out << p->info->upperBnd.getVariance();
			else *out << "----";
			*out << " lower: ";
			if (p->info->haveLower)	*out << p->info->lowerBnd;
			else *out << "----";
			*out << "\n";
			++p;						// advance
			if (++nPrinted >= MAX_TO_PRINT) {
				int remainder = lq.size() - MAX_TO_PRINT;
				if (remainder > 0) {
					*out << "  ...followed by " << lq.size() - MAX_TO_PRINT << " more...\n";
				}
				break;
			}
		}
	}
}
